import sys

from cpl_core.configuration import ConfigurationABC
from cpl_core.console import Console
from cpl_query.extension import List, IterableABC

from py_to_uxf_core.abc.umlet_creator_abc import UmletCreatorABC
from py_to_uxf_core.configuration.parser_settings import ParserSettings
from py_to_uxf_core.configuration.uml_creator_settings import UMLCreatorSettings
from py_to_uxf_core.model.class_implementation import ClassImplementation
from py_to_uxf_core.model.dimension import Dimension
from py_to_uxf_core.model.position import Position
from py_to_uxf_core.model.python_class import PythonClass
from py_to_uxf_core.model.uml_class import UMLClass

OrderedImplementation = dict[ClassImplementation, IterableABC]


class UmletCreatorService(UmletCreatorABC):

    def __init__(self, config: ConfigurationABC):
        UmletCreatorABC.__init__(self)

        self._settings: UMLCreatorSettings = config.get_configuration(UMLCreatorSettings)
        self._parser_settings: ParserSettings = config.get_configuration(ParserSettings)
        self._moved_classes: list[UMLClass] = []
        self._padding = 30

    def _sort_by_implementation(self, uml_classes: List[UMLClass], implementations: List[ClassImplementation]):
        base_sub_map: dict[UMLClass, list[UMLClass]] = {}
        for implementation in implementations:
            sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
            base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()

            if base_class not in base_sub_map:
                base_sub_map[base_class] = []

            if sub_class in base_sub_map[base_class]:
                continue
            base_sub_map[base_class].append(sub_class)

        last_base = Position(self._padding, self._padding)
        highest_y = self._padding

        for base in base_sub_map:
            subclasses = base_sub_map[base]

            base.position.x = last_base.x
            base.position.y = last_base.y

            last_sub = Position(base.position.x, base.position.y + base.dimension.height)
            for sub_class in subclasses:
                if sub_class not in self._moved_classes:
                    sub_class.position.x = base.position.x
                    sub_class.position.y = base.position.y

                if sub_class.position.x <= last_sub.x:
                    sub_class.position.x = last_sub.x
                if sub_class.position.y <= last_sub.y:
                    sub_class.position.y = last_sub.y + self._padding * 2

                last_sub.x = sub_class.position.x + sub_class.dimension.width + self._padding
                if sub_class.position.y + sub_class.dimension.height > highest_y:
                    highest_y = sub_class.position.y + sub_class.dimension.height

                self._moved_classes.append(sub_class)

            last_base.x = last_sub.x
            if last_base.x <= self._settings.max_pixel_x:
                last_base.x += self._padding * 2
            else:
                last_base.x = self._padding
                last_base.y = highest_y + self._padding * 2

            self._moved_classes.append(base)

        for cls in uml_classes:
            if cls in self._moved_classes:
                continue

            new_x = last_base.x
            if new_x <= self._settings.max_pixel_x:
                cls.position.x = new_x
                last_base.x = new_x + cls.dimension.width + self._padding
                cls.position.y = last_base.y
            else:
                cls.position.x = self._padding
                cls.position.y = highest_y + self._padding * 2
                last_base = cls.position

            if cls.position.y + cls.dimension.height > highest_y:
                highest_y = cls.position.y + cls.dimension.height

    def _sort_implementations_by_base_order(self, implementations: List[ClassImplementation]) -> OrderedImplementation:
        """
        Sort given implementations by order of base classes
        SubClass(FirstBaseClass, SecondBaseClass, ThirdBaseClass)
        """
        ordered_implementations: OrderedImplementation = {}
        for implementation in list(implementations):
            if not implementation.is_first or implementation.base.name in self._parser_settings.ignore_class_names:
                continue

            ordered_implementations[implementation] = implementations.where(lambda i: i.subclass == implementation.subclass and not i.is_first)

        return ordered_implementations

    def _move_by_implementation(self, implementations: OrderedImplementation, uml_classes: List[UMLClass]) -> List[UMLClass]:
        base_sub_map: dict[UMLClass, list[UMLClass]] = {}
        for implementation in implementations:
            sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
            base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()

            if base_class not in base_sub_map:
                base_sub_map[base_class] = []

            if sub_class in base_sub_map[base_class]:
                continue

            base_sub_map[base_class].append(sub_class)
            for sub_impl in implementations[implementation]:
                si_sub_class: UMLClass = uml_classes.where(lambda u: u.cls == sub_impl.subclass).first()
                si_base_class: UMLClass = uml_classes.where(lambda u: u.cls == sub_impl.base).first()

                if si_base_class not in base_sub_map:
                    base_sub_map[si_base_class] = []

                if si_sub_class in base_sub_map[si_base_class]:
                    continue
                base_sub_map[si_base_class].append(si_sub_class)

        moved_classes = List(UMLClass)
        next_pos = Position(self._padding, self._padding)
        highest_y = 0
        for base in base_sub_map:
            subclasses = base_sub_map[base]
            if base in moved_classes:
                Console.write_line('b_SKIP', base)
                continue

            if next_pos.x >= self._settings.max_pixel_x:
                next_pos.x = self._padding
                next_pos.y = highest_y + self._padding

            base.position.x = next_pos.x
            base.position.y = next_pos.y
            if base.position.y + base.dimension.height > highest_y:
                highest_y = base.position.y + base.dimension.height

            next_pos.y = base.position.y + base.dimension.height + self._padding

            Console.write_line('moved b', base)
            moved_classes.append(base)

            moved_subclasses = []
            for sub in subclasses:
                if sub in moved_classes:
                    Console.write_line('s_SKIP', sub)
                    continue

                sub.position.x = next_pos.x
                sub.position.y = next_pos.y
                next_pos.x += sub.dimension.width + self._padding
                if sub.position.y + sub.dimension.height > highest_y:
                    highest_y = sub.position.y + sub.dimension.height

                Console.write_line('moved s', sub)
                moved_classes.append(sub)
                moved_subclasses.append(sub)

            delta_x = self._padding
            if len(moved_subclasses) == 0:
                delta_x += base.dimension.width
            next_pos.x += delta_x
            next_pos.y = base.position.y

        for cls in uml_classes:
            if cls in moved_classes:
                continue

            new_x = next_pos.x
            if new_x <= self._settings.max_pixel_x:
                cls.position.x = new_x
                next_pos.x = new_x + cls.dimension.width + self._padding
                cls.position.y = next_pos.y
            else:
                cls.position.x = self._padding
                cls.position.y = highest_y + self._padding * 2
                next_pos.x = cls.position.x
                next_pos.y = cls.position.y

            if cls.position.y + cls.dimension.height > highest_y:
                highest_y = cls.position.y + cls.dimension.height

            moved_classes.append(cls)
            # Console.write_line('moved c', cls)

        return moved_classes

    def generate_xml(self, classes: List[PythonClass], implementations: List[ClassImplementation]) -> str:
        uml_classes = List(UMLClass)
        xml_classes = ''
        xml_relations = ''

        for cls in classes:
            uml_cls = UMLClass(cls, Position(0, self._padding), Dimension(80, 80))
            # trigger xml generation to render coordinates
            uml_cls.create_xml_body()
            uml_classes.append(uml_cls)

        # self._sort_by_implementation(uml_classes, implementations)
        oi = self._sort_implementations_by_base_order(implementations)
        for uml_cls in self._move_by_implementation(oi, uml_classes):
            # save class xml
            if uml_cls.cls.name in self._parser_settings.ignore_class_names:
                continue

            xml_classes += uml_cls.as_xml()

        for implementation in implementations:
            sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
            base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()
            if base_class.cls.name in self._parser_settings.ignore_class_names:
                continue
            xml_relations += sub_class.implementation_as_xml(base_class)

        return f"""\
        <?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <diagram program="umlet" version="14.1.1">
            <zoom_level>10</zoom_level>
            {xml_classes}
            {xml_relations}
        </diagram>
        """.replace('    ', '').replace('\t', '')
